from robot.libraries.BuiltIn import BuiltIn

from testit_python_commons.services import TmsPluginManager

from .listeners import AutotestAdapter, TestRunAdapter
from .models import Option


def enabled(func):
    def wrapped(self, *args, **kwargs):
        if self.enabled:
            return func(self, *args, **kwargs)
        else:
            raise ImportError("TestIt module should be enabled. Use '-v testit' CLI option")

    return wrapped


class TMSLibrary:
    """Library for exporting result to TestIt.

        = Table of contents =

        %TOC%

        = Usage =

        This library has several keyword, for example `Add Link`, adding links to result of test in TestIt

        = Examples =

        | `Add Message`      | My message    |                |               |
        | `Add Link`         | http://ya.ru  |                |               |
        | `Add Attachments`  | image.png     | log.txt        | video.gif     |
        """
    ROBOT_LIBRARY_SCOPE = 'GLOBAL'
    ROBOT_LIBRARY_VERSION = '1.0'

    def __init__(self):
        built_in = BuiltIn()
        self.enabled = built_in.get_variable_value("${testit}", None) is not None
        if self.enabled:
            cli_params = ["tmsUrl", "tmsPrivateToken", "tmsProjectId", "tmsConfigurationId", "tmsTestRunId",
                          "tmsProxy", "tmsTestRunName", "tmsAdapterMode", "tmsConfigFile", "tmsCertValidation",
                          "tmsAutomaticCreationTestCases", "tmsAutomaticUpdationLinksToTestCases", "tmsImportRealtime"]
            option = Option(**{param: built_in.get_variable_value(f'${{{param}}}', None) for param in cli_params})
            self.adapter_manager = TmsPluginManager.get_adapter_manager(option)
            pabot_index = built_in.get_variable_value('${PABOTQUEUEINDEX}', None)
            if pabot_index is not None:
                try:
                    from pabot import PabotLib
                    pabot = PabotLib()
                    if int(pabot_index) == 0:
                        test_run_id = self.adapter_manager.get_test_run_id()
                        pabot.set_parallel_value_for_key('test_run_id', test_run_id)
                    else:
                        while True:
                            test_run_id = pabot.get_parallel_value_for_key('test_run_id')
                            if test_run_id:
                                break
                    self.adapter_manager.set_test_run_id(test_run_id)
                except RuntimeError:
                    raise SystemExit
            else:
                self.adapter_manager.set_test_run_id(self.adapter_manager.get_test_run_id())
            self.ROBOT_LIBRARY_LISTENER = [AutotestAdapter(self.adapter_manager), TestRunAdapter(self.adapter_manager)]

    @enabled
    def add_link(self, url, type='Defect', title=None, description=None):  # noqa: A002,VNE003
        """
        Adds link to current test.

        Valid link types are ``Defect``, ``Issue``, ``Related``, ``BlockedBy``, ``Requirement``, ``Repository``.

        """
        from testit_python_commons.models.link import Link

        link = Link()\
            .set_url(url)\
            .set_title(title)\
            .set_link_type(type)\
            .set_description(description)
        self.ROBOT_LIBRARY_LISTENER[0].active_test.resultLinks.append(link)

    @enabled
    def add_links(self, *links):
        """
        Adds several links to current test.

        Every link should be a dict with ``url`` key. See `Add Link` keyword for more information.

        """
        for link in links:
            if isinstance(link, dict):
                self.add_link(**link)

    @enabled
    def add_attachments(self, *paths):
        """
        Adds several attachments to current test.

        """
        attachments = self.adapter_manager.load_attachments(paths)
        self.ROBOT_LIBRARY_LISTENER[0].active_test.attachments.extend(attachments)

    @enabled
    def add_attachment(self, text, filename=None):
        """
        Adds attachment to current test

        """
        attachment = self.adapter_manager.create_attachment(text, filename)
        self.ROBOT_LIBRARY_LISTENER[0].active_test.attachments.extend(attachment)

    @enabled
    def add_message(self, message):
        """
        Adds error message to current test
        """
        self.ROBOT_LIBRARY_LISTENER[0].active_test.message = message



import ast
import re

from attr import Factory, asdict, attrib, s

from robot.api import logger
from testit_python_commons.models.link import Link

from .utils import get_hash


LinkTypes = ['Related', 'BlockedBy', 'Defect', 'Issue', 'Requirement', 'Repository']


def link_type_check(self, attribute, value):
    if value.title() not in LinkTypes:
        raise ValueError(f"Incorrect Link type: {value}")


def url_check(self, attribute, value):
    if not bool(re.match(
            r"(https?|ftp)://"
            r"(\w+(-\w+)*\.)?"
            r"((\w+(-\w+)*)\.(\w+))"
            r"(\.\w+)*"
            r"([\w\-._~/]*)*(?<!\.)",
            value)):
        raise ValueError(f"Incorrect URL: {value}")


class Default:

    def order(self):
        return asdict(self)


@s
class StepResult(Default):
    title = attrib(default='')
    description = attrib(default='')
    started_on = attrib(default=None)
    completed_on = attrib(default=None)
    duration = attrib(default=None)
    outcome = attrib(default=None)
    step_results = attrib(default=Factory(list))
    attachments = attrib(default=Factory(list))
    parameters = attrib(default=Factory(dict))


@s
class Step(Default):
    title = attrib()
    description = attrib()
    steps = attrib(default=Factory(list))


@s
class Label:
    name = attrib()


@s(kw_only=True)
class Autotest(Default):
    externalID = attrib(default=None)  # noqa: N815
    autoTestName = attrib()  # noqa: N815
    steps = attrib(default=Factory(list))
    stepResults = attrib(default=Factory(list))  # noqa: N815
    setUp = attrib(default=Factory(list))  # noqa: N815
    setUpResults = attrib(default=Factory(list))  # noqa: N815
    tearDown = attrib(default=Factory(list))  # noqa: N815
    tearDownResults = attrib(default=Factory(list))  # noqa: N815
    resultLinks = attrib(default=Factory(list))  # noqa: N815
    duration = attrib(default=None)
    failureReasonNames = attrib(default=Factory(list))  # noqa: N815
    traces = attrib(default=None)
    outcome = attrib(default=None)
    namespace = attrib(default=None)
    attachments = attrib(default=Factory(list))
    parameters = attrib(default=Factory(dict))
    properties = attrib(default=Factory(dict))
    classname = attrib(default=None)
    title = attrib(default=None)
    description = attrib(default=None)
    links = attrib(default=Factory(list))
    labels = attrib(default=Factory(list))
    workItemsID = attrib(default=Factory(list))  # noqa: N815
    message = attrib(default="")
    started_on = attrib(default=None)
    completed_on = attrib(default=None)
    externalKey = attrib(default=None)

    step_depth = attrib(default=Factory(list))
    result_depth = attrib(default=Factory(list))

    def add_attributes(self, attrs):
        self.title = attrs['originalname']
        self.autoTestName = attrs['originalname']
        self.externalKey = attrs['originalname']
        self.description = attrs['doc']
        self.template = attrs['template']
        self.classname = attrs['longname'].split('.')[-2]
        for tag in attrs['tags']:
            if tag.lower().startswith('testit.'):
                attr = re.findall(r'(?<=\.).*?(?=:)', tag)[0].strip().lower()
                value = tag.split(':', 1)[-1].strip()
                if attr == 'externalid':
                    self.externalID = str(value).replace("'", "").replace('"', '')
                elif attr == 'displayname':
                    self.autoTestName = str(value).replace("'", "").replace('"', '')
                elif attr == 'title':
                    self.title = str(value).replace("'", "").replace('"', '')
                elif attr == 'description':
                    self.description = str(value).replace("'", "").replace('"', '')
                elif attr == 'workitemsid' or attr == 'workitemsids':
                    value = ast.literal_eval(value)
                    if isinstance(value, (str, int)):
                        self.workItemsID.append(str(value))
                    elif isinstance(value, list):
                        self.workItemsID.extend([str(i) for i in value])
                    else:
                        logger.error(f"[TestIt] Wrong workitem format: {value}")
                elif attr == 'links':
                    value = ast.literal_eval(value)
                    try:
                        if isinstance(value, dict):
                            self.links.append(Link()\
                                .set_url(value['url'])\
                                .set_title(value.get('title', None))\
                                .set_link_type(value.get('type', None))\
                                .set_description(value.get('description', None)))
                        elif isinstance(value, list):
                            self.links.extend([Link()\
                                .set_url(link['url'])\
                                .set_title(link.get('title', None))\
                                .set_link_type(link.get('type', None))\
                                .set_description(link.get('description', None)) for link in value if isinstance(link, dict)])
                    except ValueError as e:
                        logger.error(f"[TestIt] Link Error: {e}")
                elif attr == 'labels':
                    value = ast.literal_eval(value)
                    if isinstance(value, (str, int)):
                        self.labels.append(Label(value))
                    elif isinstance(value, list):
                        self.labels.extend([Label(item) for item in value if isinstance(item, (str, int))])
                elif attr == 'namespace':
                    self.namespace = str(value).replace("'", "").replace('"', '')
                elif attr == 'classname':
                    self.classname = str(value).replace("'", "").replace('"', '')
                else:
                    logger.error(f"[TestIt] Unknown attribute: {attr}")
        if not self.externalID:
            self.externalID = get_hash(attrs['longname'])

    def add_step(self, step_type, title, description, parameters):
        if len(self.step_depth) == 0:
            if step_type.lower() == 'setup':
                self.setUp.append(Step(title, description))
                self.step_depth.append(self.setUp[-1])
                self.setUpResults.append(StepResult(title, description, parameters=parameters))
                self.result_depth.append(self.setUpResults[-1])
            elif step_type.lower() == 'teardown':
                self.tearDown.append(Step(title, description))
                self.step_depth.append(self.tearDown[-1])
                self.tearDownResults.append(StepResult(title, description, parameters=parameters))
                self.result_depth.append(self.tearDownResults[-1])
            else:
                self.steps.append(Step(title, description))
                self.step_depth.append(self.steps[-1])
                self.stepResults.append(StepResult(title, description, parameters=parameters))
                self.result_depth.append(self.stepResults[-1])
        elif 1 <= len(self.step_depth) < 14:
            self.step_depth[-1].steps.append(Step(title, description))
            self.step_depth.append(self.step_depth[-1].steps[-1])
            self.result_depth[-1].step_results.append(StepResult(title, description, parameters=parameters))
            self.result_depth.append(self.result_depth[-1].step_results[-1])

    def add_step_result(self, title, start, complete, duration, outcome, attachments):
        if self.result_depth:
            if self.result_depth[-1].title == title:
                step = self.result_depth.pop()
                step.started_on = start
                step.completed_on = complete
                step.duration = duration
                step.outcome = outcome
                step.attachments = attachments
        if self.step_depth:
            if self.step_depth[-1].title == title:
                self.step_depth.pop()


class Option:

    def __init__(self, **kwargs):
        if kwargs.get('tmsUrl', None):
            self.set_url = kwargs.get('tmsUrl', None)
        if kwargs.get('tmsPrivateToken', None):
            self.set_private_token = kwargs.get('tmsPrivateToken', None)
        if kwargs.get('tmsProjectId', None):
            self.set_project_id = kwargs.get('tmsProjectId', None)
        if kwargs.get('tmsConfigurationId', None):
            self.set_configuration_id = kwargs.get('tmsConfigurationId', None)
        if kwargs.get('tmsTestRunId', None):
            self.set_test_run_id = kwargs.get('tmsTestRunId', None)
        if kwargs.get('tmsProxy', None):
            self.set_tms_proxy = kwargs.get('tmsProxy', None)
        if kwargs.get('tmsTestRunName', None):
            self.set_test_run_name = kwargs.get('tmsTestRunName', None)
        if kwargs.get('tmsAdapterMode', None):
            self.set_adapter_mode = kwargs.get('tmsAdapterMode', None)
        if kwargs.get('tmsConfigFile', None):
            self.set_config_file = kwargs.get('tmsConfigFile', None)
        if kwargs.get('tmsCertValidation', None):
            self.set_cert_validation = kwargs.get('tmsCertValidation', None)
        if kwargs.get('tmsAutomaticCreationTestCases', None):
            self.set_automatic_creation_test_cases = kwargs.get('tmsAutomaticCreationTestCases', None)
        if kwargs.get('tmsAutomaticUpdationLinksToTestCases', None):
            self.set_automatic_updation_links_to_test_cases = kwargs.get('tmsAutomaticUpdationLinksToTestCases', None)
        if kwargs.get('tmsImportRealtime', None):
            self.set_import_realtime = kwargs.get('tmsImportRealtime', None)



import hashlib
from datetime import datetime
from typing import List

from testit_python_commons.models.step_result import StepResult
from testit_python_commons.models.test_result import TestResult


def convert_time(time):
    date = datetime.strptime(time, "%Y%m%d %H:%M:%S.%f")
    date = date.replace(tzinfo=datetime.now().astimezone().tzinfo)
    return date


def convert_executable_test_to_test_result_model(executable_test: dict) -> TestResult:
    return TestResult()\
        .set_external_id(executable_test['externalID'])\
        .set_autotest_name(executable_test['autoTestName'])\
        .set_step_results(
            step_results_to_autotest_steps_model(executable_test['stepResults']))\
        .set_setup_results(
            step_results_to_autotest_steps_model(executable_test['setUpResults']))\
        .set_teardown_results(
            step_results_to_autotest_steps_model(executable_test['tearDownResults']))\
        .set_duration(executable_test['duration'])\
        .set_outcome(executable_test['outcome'])\
        .set_traces(executable_test['traces'])\
        .set_attachments(executable_test['attachments'])\
        .set_parameters(executable_test['parameters'])\
        .set_properties(executable_test['properties'])\
        .set_namespace(executable_test['namespace'])\
        .set_classname(executable_test['classname'])\
        .set_title(executable_test['title'])\
        .set_description(executable_test['description'])\
        .set_links(executable_test['links'])\
        .set_result_links(executable_test['resultLinks'])\
        .set_labels(executable_test['labels'])\
        .set_work_item_ids(executable_test['workItemsID'])\
        .set_message(executable_test['message'])\
        .set_external_key(executable_test['externalKey'])


def step_results_to_autotest_steps_model(step_results: dict) -> List[StepResult]:
    autotest_model_steps = []

    for step_result in step_results:
        step_result_model = StepResult()\
            .set_title(step_result['title'])\
            .set_description(step_result['description'])\
            .set_outcome(step_result['outcome'])\
            .set_duration(step_result['duration'])\
            .set_attachments(step_result['attachments'])

        if 'parameters' in step_result:
            step_result_model.set_parameters(step_result['parameters'])

        if 'step_results' in step_result:
            step_result_model.set_step_results(
                step_results_to_autotest_steps_model(step_result['step_results']))

        autotest_model_steps.append(step_result_model)

    return autotest_model_steps


def get_hash(value: str):
    md = hashlib.sha256(bytes(value, encoding='utf-8'))
    return md.hexdigest()


STATUSES = {'FAIL': 'Failed', 'PASS': 'Passed', 'SKIP': 'Skipped', 'NOT RUN': 'Skipped'}



import re

from robot.api import SuiteVisitor, logger
from robot.libraries.BuiltIn import BuiltIn

from .models import Autotest
from .utils import STATUSES, convert_time, convert_executable_test_to_test_result_model, get_hash


class AutotestAdapter:
    ROBOT_LISTENER_API_VERSION = 2

    def __init__(self, adapter_manager):
        self.adapter_manager = adapter_manager
        self.active_test = None

    @staticmethod
    def get_test_title(attrs):
        title = attrs['type']
        return "\t".join(
            attrs['assign'] + [title if title not in ["SETUP", "TEARDOWN", "KEYWORD"] else "",
                               attrs['kwname']] + attrs['args'])

    @staticmethod
    def parse_arguments(args):
        parameters = {}
        for arg in args:
            variables = re.findall(r'\${[a-zA-Z-_\\ \d]*}', arg)
            for arg_var in variables:
                value = str(BuiltIn().get_variable_value(arg_var))
                if len(value) > 2000:
                    value = value[:2000]
                parameters[arg_var] = value
        return parameters if parameters else None

    def start_test(self, name, attributes):
        self.active_test = Autotest(autoTestName=name)
        self.active_test.add_attributes(attributes)
        BuiltIn().remove_tags("testit*")

    def start_keyword(self, name, attributes):
        if self.active_test:
            title = self.get_test_title(attributes)
            parameters = self.parse_arguments(attributes['args'])
            self.active_test.add_step(attributes['type'], title, attributes['doc'], parameters)

    def end_keyword(self, name, attributes):
        if self.active_test:
            title = self.get_test_title(attributes)
            start = convert_time(attributes['starttime'])
            end = convert_time(attributes['endtime'])
            duration = attributes['elapsedtime']
            outcome = STATUSES[attributes['status']]
            self.active_test.add_step_result(title, start, end, duration, outcome, self.active_test.attachments)
            self.active_test.attachments = []

    def end_test(self, name, attributes):
        if self.active_test:
            self.active_test.outcome = STATUSES[attributes['status']]
            self.active_test.started_on = convert_time(attributes['starttime'])
            self.active_test.completed_on = convert_time(attributes['endtime'])
            if not self.active_test.message:
                if self.active_test.outcome == 'Failed':
                    for step in (self.active_test.setUpResults + self.active_test.stepResults +
                                 self.active_test.tearDownResults):
                        if step.outcome == 'Failed':
                            self.active_test.message = f"Failed on step: '{step.title}'"
                            break
            self.active_test.traces = attributes['message']
            self.active_test.duration = attributes['elapsedtime']
            self.adapter_manager.write_test(
                convert_executable_test_to_test_result_model(self.active_test.order()))

    def close(self):
        self.adapter_manager.write_tests()


class TestRunAdapter:
    ROBOT_LISTENER_API_VERSION = 3

    def __init__(self, adapter_manager):
        self.adapter_manager = adapter_manager

    def start_suite(self, suite, result):
        tests = self.adapter_manager.get_autotests_for_launch()
        if tests is not None:
            selector = ExcludeTests(*tests)
            suite.visit(selector)


class ExcludeTests(SuiteVisitor):

    def __init__(self, *tests):
        self.tests = tests
        if not len(self.tests):
            logger.error('No tests to run!')
            raise SystemExit

    def start_suite(self, suite):
        suite.tests = [t for t in suite.tests if self._is_included(t)]

    def _is_included(self, test):
        tags = test.tags
        external_id = get_hash(test.longname)
        for tag in tags:
            if str(tag).lower().startswith('testit.externalid'):
                external_id = tag.split(':', 1)[-1].strip()
        return external_id in self.tests

    def end_suite(self, suite):
        suite.suites = [s for s in suite.suites if s.test_count > 0]

    def visit_test(self, test):
        pass



import re
from pytest_mock import MockerFixture
from unittest.mock import MagicMock


class TestAutotestAdapter:
    def test_parse_arguments(self, mocker: MockerFixture):
        mock_builtin_class = mocker.Mock()
        mock_builtin_instance = mocker.Mock()
        mock_builtin_class.return_value = mock_builtin_instance
        
        mock_builtin_instance.get_variable_value.side_effect = [
            "value1",
            "value2",
            "a" * 2500  # A value longer than 2000 characters
        ]
        
        def parse_arguments(args):
            parameters = {}
            for arg in args:
                variables = re.findall(r'\${[a-zA-Z-_\\ \d]*}', arg)
                for arg_var in variables:
                    value = str(mock_builtin_instance.get_variable_value(arg_var))
                    if len(value) > 2000:
                        value = value[:2000]
                    parameters[arg_var] = value
            return parameters if parameters else None
        
        # Arrange
        args = ["${var1}", "prefix_${var2}_suffix", "${var3_with_long_value}"]
        
        # Act
        parameters = parse_arguments(args)
        
        # Assert
        assert parameters == {
            "${var1}": "value1",
            "${var2}": "value2",
            "${var3_with_long_value}": "a" * 2000  # Should be truncated
        }
        
        expected_calls = [
            mocker.call("${var1}"),
            mocker.call("${var2}"),
            mocker.call("${var3_with_long_value}")
        ]
        mock_builtin_instance.get_variable_value.assert_has_calls(expected_calls, any_order=True)